Attribute VB_Name = "ImageResizer"
'###############################################################
'# Coded By Armin Niki ( Nickname ) for Photormin Open Source  #\
'# ( Supported by SourceForge )                        Project #\
'####### E-mail to Armin3D@gmail.com For More Information ######\
' \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

Option Explicit

Type RECT
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
End Type

Type GUID
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(7) As Byte
End Type

Type PicBmp
    Size As Long
    Type As PictureTypeConstants
    hBmp As Long
    hPal As Long
    Reserved As Long
End Type

' Representation of 32-bit RGBA color
Type RGBQUAD
    rgbRed As Byte
    rgbGreen As Byte
    rgbBlue As Byte
    rgbReserved As Byte
End Type

Type BITMAPINFOHEADER
    biSize As Long
    biWidth As Long
    biHeight As Long
    biPlanes As Integer
    biBitCount As Integer
    biCompression As Long
    biSizeImage As Long
    biXPelsPerMeter As Long
    biYPelsPerMeter As Long
    biClrUsed As Long
    biClrImportant As Long
End Type

Type BITMAPINFO
    bmiHeader As BITMAPINFOHEADER
    bmiColors As RGBQUAD
End Type

Type BITMAP
    bmType As Long
    bmWidth As Long
    bmHeight As Long
    bmWidthBytes As Long
    bmPlanes As Integer
    bmBitsPixel As Integer
    bmBits As Long
End Type

Type SAFEARRAYBOUND
    cElements As Long
    lLbound As Long
End Type

' VB's array header structure
Type SAFEARRAY2D
    cDims As Integer
    fFeatures As Integer
    cbElements As Long
    cLocks As Long
    pvData As Long
    Bounds(1) As SAFEARRAYBOUND
End Type

' Some API calls

' Copies memory
Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (lpvDest As Any, lpvSource As Any, ByVal cbCopy&)

' Draws text to device
Declare Function DrawText Lib "user32" Alias "DrawTextA" (ByVal hDc&, ByVal lpStr$, ByVal nCount&, lpRect As RECT, ByVal wFormat&) As Long

' Creates device independent bitmap
Declare Function CreateDIBSection Lib "gdi32" (ByVal hDc&, pBitmapInfo As BITMAPINFO, ByVal un&, lplpVoid&, ByVal handle&, ByVal dw&) As Long
' Gets info about system object
Declare Function GetObjectAPI Lib "gdi32" Alias "GetObjectA" (ByVal hObject&, ByVal nCount&, lpObject As Any) As Long
' Creates GDI device
Declare Function CreateCompatibleDC Lib "gdi32" (ByVal hDc&) As Long
Declare Function DeleteDC Lib "gdi32" (ByVal hDc&) As Long
Declare Function DeleteObject Lib "gdi32" (ByVal hObject&) As Long
' Associates device with GDI object
Declare Function SelectObject Lib "gdi32" (ByVal hDc&, ByVal hObject&) As Long
' Copies a part of bitmap to another
Declare Function BitBlt Lib "gdi32" (ByVal hDestDC&, ByVal x&, ByVal y&, ByVal nWidth&, ByVal nHeight&, ByVal hSrcDC&, ByVal xSrc&, ByVal ySrc&, ByVal dwRop&) As Long

' Creates OLE Picture object used in VB
Declare Function OleCreatePictureIndirect Lib "olepro32" (PicDesc As PicBmp, RefIID As GUID, ByVal fPictureOwnsHandle&, IPic As IPicture) As Long

' Points to any variable (where VarPtr() doesn't work)
Declare Function VarPtrArray Lib "msvbvm60" Alias "VarPtr" (Ptr() As Any) As Long


Enum eTypes
    NearestNeighbor
    Bilinear
    BicubicCardinal
    BicubicBSpline
    BicubicBCSpline
    Bell
    Gaussian
    WindowedSinc
End Enum

Enum eWindows
    wBartlett
    wBlackman
    wBlackmanHarris
    wBohman
    wConnes
    wCosine
    wGaussian
    wHamming
    wHann
    wKaiser
    wLanczos
    wParzen
    wRectangular
    wWelch
End Enum

Const pi = 3.14159265358979

Public cubic_a#, cubic_b#, cubic_c#, param_d#
Public kernel_size&, sinc_window&, stair_level&

' Creates Picture object of given dimensions and depth
Function CreatePicture(ByVal nWidth&, ByVal nHeight&, ByVal nBPP&) As Picture
    Dim pic As PicBmp, IID_IDispatch As GUID, BMI As BITMAPINFO
    With BMI.bmiHeader
        .biSize = Len(BMI.bmiHeader)
        .biWidth = nWidth
        .biHeight = nHeight
        .biPlanes = 1
        .biBitCount = nBPP
    End With
    pic.hBmp = CreateDIBSection(0, BMI, 0, 0, 0, 0)
    With IID_IDispatch
        .Data1 = &H20400: .Data4(0) = &HC0: .Data4(7) = &H46
    End With
    pic.Size = Len(pic)
    pic.Type = vbPicTypeBitmap
    OleCreatePictureIndirect pic, IID_IDispatch, 1, CreatePicture
    If CreatePicture = 0 Then Set CreatePicture = Nothing
End Function

' Copies one memory bitmap into another with edge extending
Sub CopyImage24(InArray() As Byte, OutArray() As Byte, ByVal InUBound&, Optional ByVal DisablePad As Boolean)
    Dim i&, j&
    j = InUBound + 1
    ' Copy full content of input image into output
    For i = 0 To UBound(InArray, 2)
        CopyMemory OutArray(0, i), InArray(0, i), j
    Next
    ' Fill extended pad bytes with color of edges or not?
    If DisablePad Then Exit Sub
    ' Fill left and right edges
    For j = 0 To UBound(InArray, 2)
        For i = LBound(OutArray) To -3 Step 3
            OutArray(i, j) = OutArray(0, j) ' Blue
            OutArray(i + 1, j) = OutArray(1, j) ' Green
            OutArray(i + 2, j) = OutArray(2, j) ' Red
        Next
        For i = InUBound + 1 To UBound(OutArray) - 2 Step 3
            OutArray(i, j) = OutArray(InUBound - 2, j)
            OutArray(i + 1, j) = OutArray(InUBound - 1, j)
            OutArray(i + 2, j) = OutArray(InUBound, j)
        Next
    Next
    j = UBound(OutArray) - LBound(OutArray) + 1
    InUBound = LBound(OutArray)
    ' Fill top and bottom edges
    For i = LBound(OutArray, 2) To -1
        CopyMemory OutArray(InUBound, i), OutArray(InUBound, 0), j
    Next
    For i = UBound(InArray, 2) + 1 To UBound(OutArray, 2)
        CopyMemory OutArray(InUBound, i), OutArray(InUBound, UBound(InArray, 2)), j
    Next
End Sub


' Simple Box function
Function Box_func(ByVal x#) As Double
    If Abs(x) < 0.5 Then Box_func = 1
End Function

' Linear function with various extent
Function Bartlett_Linear(ByVal x#) As Double
    x = Abs(x)
    If x < kernel_size Then Bartlett_Linear = (1 - Abs(x) / kernel_size) / kernel_size
End Function

' Cardinal cubic spline function
' Depends on one variable: a
' These splines are derived from BC-splines with B=0 and C=-a.
' Only for a=-0.5 the Taylor series expansion of the interpolating function is optimal.
' This cardinal spline is then called Catmull-Rom spline.
' If a=0 then the spline is called Hermite.
Function Cardinal_Cubic_Spline(ByVal x#) As Double
    x = Abs(x)
    If x < 1 Then
        Cardinal_Cubic_Spline = (cubic_a + 2) * x * x * x - (cubic_a + 3) * x * x + 1
    ElseIf x < 2 Then
        Cardinal_Cubic_Spline = cubic_a * x * x * x - 5 * cubic_a * x * x + 8 * cubic_a * x - 4 * cubic_a
    End If
End Function

' Cubic B-spline function
' Depends on no variable.
' It is derived from BC-splines with B=1 and C=0.
Function Cubic_BSpline(ByVal x#) As Double
    Dim a#, b#, c#, d#, tmp#
    If x < 2 Then
        tmp = x + 2
        If tmp > 0 Then a = tmp * tmp * tmp
        tmp = x + 1
        If tmp > 0 Then b = 4 * tmp * tmp * tmp
        If x > 0 Then c = 6 * x * x * x
        tmp = x - 1
        If tmp > 0 Then d = 4 * tmp * tmp * tmp
        Cubic_BSpline = (a - b + c - d) / 6
    End If
End Function

' Cubic BC-spline function
' Mitchell and Netravali derived a family of such cubic filters dependent on two variables: B, C
' Some of the values for B and C correspond to well-known filters,
' e.g., B=1 and C=0 corresponds to the cubic B-spline,
' and C=0 results in the family of Duff's tensioned B-splines.
' Setting B=0 and C=-a results in the family of the cardinal splines which were derived by Keys in 1981.
' Using Taylor series expansion they determined that, numerically, the filters for which B + 2 * C = 1 with 0 <= B <= 1
' are the most accurate within that class
' and that the reconstruction error for synthetic examples is proportional to the square of the sampling distance.
' Two new filters were proposed, the first with B=3/2 and C=1/3 suppresses post-aliasing but is unnecessarily blurring,
' the second with B=1/3 and C=1/3 turns out to be a satisfactory compromise between ringing, blurring, and anisotropy.
Function Cubic_BCSpline(ByVal x#) As Double
    x = Abs(x)
    If x < 1 Then
        Cubic_BCSpline = ((12 - 9 * cubic_b - 6 * cubic_c) * x * x * x + (-18 + 12 * cubic_b + 6 * cubic_c) * x * x + 6 - 2 * cubic_b) / 6
    ElseIf x < 2 Then
        Cubic_BCSpline = ((-cubic_b - 6 * cubic_c) * x * x * x + (6 * cubic_b + 30 * cubic_c) * x * x + (-12 * cubic_b - 48 * cubic_c) * x + (8 * cubic_b + 24 * cubic_c)) / 6
    End If
End Function

' Bell function
Function Bell_Func(ByVal x#) As Double
    x = Abs(x)
    If x < 0.5 Then
        Bell_Func = 0.75 - x * x
    ElseIf x < 1.5 Then
        Bell_Func = 0.5 * (x - 1.5) * (x - 1.5)
    End If
End Function

' Gaussian function
' Could generate very blurry output.
' The wider the function is (higher standard deviation),
' more blurry the image is. This is useful for removing
' noise and aliasing but it won't preserve details.
Function Gaussian_Func(ByVal x#) As Double
    ' 0.398942280401433 = 1 / Sqr(2 * pi)
    Dim o#
    If Abs(x) < kernel_size Then
        o = kernel_size / pi ' standard deviation - could be changed
        Gaussian_Func = 0.398942280401433 / o * EXP(-x * x / (o * o * 2))
    End If
End Function

' Windowed Sinc function
' There's a whole class of reconstruction filters created by starting out
' with the assumption that the sinc function:
'    Sinc(X) = Sin(pi * X) / (pi * X)
' is the perfect image reconstruction filter. It provides the best
' retention of the frequencies you want, and the best attenuation of the
' frequencies that you don't want (because they would cause aliasing).
' Unfortunately, it's impossible to use directly because it is infinite in extent.
' So, practical filters are created by taking the sinc function and
' multiplying it by a "window" function that gradually tapers the Sinc
' function to zero, giving an overall filter with finite size.
Function Windowed_Sinc(ByVal x#) As Double
    If x = 0 Then
        Windowed_Sinc = 1
    Else
        Windowed_Sinc = Window_Func(x) * Sin(pi * x) / (pi * x)
    End If
End Function

' Finite Impulse Response filters (FIR)
Function Window_Func(ByVal x#) As Double
    Dim i#
    If x = 0 Then
        Window_Func = 1
    ElseIf Abs(x) < kernel_size Then
        Select Case sinc_window
            Case wBartlett
                Window_Func = 1 - Abs(x) / kernel_size
            Case wBlackman
                i = pi * x / kernel_size
                Window_Func = 0.42 + 0.5 * Cos(i) + 0.08 * Cos(2 * i)
            Case wBlackmanHarris
                i = pi * x / kernel_size
                Window_Func = 0.42323 + 0.49755 * Cos(i) + 0.07922 * Cos(2 * i)
            Case wBohman
                i = x / kernel_size
                Window_Func = (1 - Abs(i)) * Cos(pi * i) + 1 / pi * Sin(pi * Abs(i))
            Case wConnes
                i = 1 - x * x / (kernel_size * kernel_size)
                Window_Func = i * i
            Case wCosine
                Window_Func = Cos(x * pi / (kernel_size * 2))
            Case wGaussian
                i = kernel_size * 0.4
                Window_Func = 2 ^ (-x * x / (i * i))
            Case wHamming
                Window_Func = 0.54 + 0.46 * Cos(pi * x / kernel_size)
            Case wHann
                Window_Func = 0.5 + 0.5 * Cos(pi * x / kernel_size)
            Case wKaiser
                Window_Func = BesselI0(param_d * Sqr(1 - x * x / (kernel_size * kernel_size))) / BesselI0(param_d)
            Case wLanczos
                i = pi * x / kernel_size
                Window_Func = Sin(i) / i
            Case wParzen
                i = Abs(x) * 1.8 / kernel_size
                If i < 1 Then
                    Window_Func = (4 - 6 * i * i + 3 * i * i * i) * 0.25
                ElseIf i < 2 Then
                    i = 2 - i
                    Window_Func = i * i * i * 0.25
                End If
            Case wRectangular
                Window_Func = 1
            Case wWelch
                Window_Func = 1 - x * x / (kernel_size * kernel_size)
        End Select
    End If
End Function

' Zero Order Modified Bessel Function of the First Kind
Function BesselI0(ByVal x#) As Double
    Dim i#
    ' The expansion usable up to x=20
    i = x * x ' X ^ 2
    BesselI0 = 1 + 0.25 * i
    i = i * i ' X ^ 4
    BesselI0 = BesselI0 + 0.015625 * i
    i = i * x * x ' X ^ 6
    BesselI0 = BesselI0 + 4.34027777777778E-04 * i
    i = i * x * x ' X ^ 8
    BesselI0 = BesselI0 + 6.78168402777778E-06 * i
    i = i * x * x ' X ^ 10
    BesselI0 = BesselI0 + 6.78168402777778E-08 * i
    i = i * x * x ' X ^ 12
    BesselI0 = BesselI0 + 4.7095027970679E-10 * i
    i = i * x * x ' X ^ 14
    BesselI0 = BesselI0 + 2.40280754952444E-12 * i
    i = i * x * x ' X ^ 16
    BesselI0 = BesselI0 + 9.38596699032984E-15 * i
    i = i * x * x ' X ^ 18
    BesselI0 = BesselI0 + 2.89690339207711E-17 * i
    i = i * x * x ' X ^ 20
    BesselI0 = BesselI0 + 7.24225848019278E-20 * i
    i = i * x * x ' X ^ 22
    BesselI0 = BesselI0 + 1.49633439673405E-22 * i
    i = i * x * x ' X ^ 24
    BesselI0 = BesselI0 + 2.59780277210772E-25 * i
    i = i * x * x ' X ^ 26
    BesselI0 = BesselI0 + 3.84290350903508E-28 * i
    i = i * x * x ' X ^ 28
    BesselI0 = BesselI0 + 4.90166263907536E-31 * i
End Function

Sub Resample_NearestNeighbor(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, kX#, kY#
    ' Compute scales on X and Y axes
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    ' Go through destination lines
    For y = dstHeight - 1 To 0 Step -1
        Y1 = y * kY ' Nearest value (rounded to integer position)
        ' Go through destination points (pixels)
        For x = 0 To dstWidth - 1
            X1 = x * kX ' Nearest value
            X1 = X1 * 3 ' 24-bit = 3 bytes
            ' Read nearest point from source and set destination point
            bDibD(x * 3, y) = bDibS(X1, Y1) ' Blue
            bDibD(x * 3 + 1, y) = bDibS(X1 + 1, Y1) ' Green
            bDibD(x * 3 + 2, y) = bDibS(X1 + 2, Y1) ' Red
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_Bilinear(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY ' Exact position (floating-point number)
        Y1 = Int(fY) ' Integer position (integer part of number)
        fY = fY - Y1 ' Fraction part of number (integer+fraction=exact)
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            ' Uses various kernel size
            For M = -kernel_size + 1 To kernel_size
                r1 = Bartlett_Linear(M - fY)
                For N = -kernel_size + 1 To kernel_size
                    r2 = Bartlett_Linear(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            bDibD(x * 3, y) = iB
            bDibD(x * 3 + 1, y) = iG
            bDibD(x * 3 + 2, y) = iR
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_BicubicCardinal(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY
        Y1 = Int(fY)
        fY = fY - Y1
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            ' Computes 1 point from 16 (4x4) surrounding points.
            For M = -1 To 2
                ' Parameter for cubic functions is distance of surrounding point and fraction part of number.
                r1 = Cardinal_Cubic_Spline(M - fY)
                For N = -1 To 2
                    r2 = Cardinal_Cubic_Spline(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            If iB < 0 Then iB = 0 ' Check possible bad RGB values
            If iG < 0 Then iG = 0
            If iR < 0 Then iR = 0
            If iB > 255 Then iB = 255
            If iG > 255 Then iG = 255
            If iR > 255 Then iR = 255
            bDibD(x * 3, y) = iB
            bDibD(x * 3 + 1, y) = iG
            bDibD(x * 3 + 2, y) = iR
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_BicubicBSpline(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY
        Y1 = Int(fY)
        fY = fY - Y1
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            For M = -1 To 2
                r1 = Cubic_BSpline(M - fY)
                For N = -1 To 2
                    r2 = Cubic_BSpline(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            bDibD(x * 3, y) = iB     ' We have no need to check values for this filter
            bDibD(x * 3 + 1, y) = iG ' because it is blurry (averaging).
            bDibD(x * 3 + 2, y) = iR ' This is also main disadvantage of B-splines.
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_BicubicBCSpline(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY
        Y1 = Int(fY)
        fY = fY - Y1
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            For M = -1 To 2
                r1 = Cubic_BCSpline(M - fY)
                For N = -1 To 2
                    r2 = Cubic_BCSpline(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            If iB < 0 Then iB = 0
            If iG < 0 Then iG = 0
            If iR < 0 Then iR = 0
            If iB > 255 Then iB = 255
            If iG > 255 Then iG = 255
            If iR > 255 Then iR = 255
            bDibD(x * 3, y) = iB
            bDibD(x * 3 + 1, y) = iG
            bDibD(x * 3 + 2, y) = iR
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_Bell(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY
        Y1 = Int(fY)
        fY = fY - Y1
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            For M = -1 To 2
                r1 = Bell_Func(M - fY)
                For N = -1 To 2
                    r2 = Bell_Func(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            bDibD(x * 3, y) = iB
            bDibD(x * 3 + 1, y) = iG
            bDibD(x * 3 + 2, y) = iR
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_Gaussian(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY
        Y1 = Int(fY)
        fY = fY - Y1
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            ' Uses various kernel size
            For M = -kernel_size + 1 To kernel_size
                r1 = Gaussian_Func(M - fY)
                For N = -kernel_size + 1 To kernel_size
                    r2 = Gaussian_Func(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            If iB < 0 Then iB = 0
            If iG < 0 Then iG = 0
            If iR < 0 Then iR = 0
            If iB > 255 Then iB = 255
            If iG > 255 Then iG = 255
            If iR > 255 Then iR = 255
            bDibD(x * 3, y) = iB
            bDibD(x * 3 + 1, y) = iG
            bDibD(x * 3 + 2, y) = iR
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

Sub Resample_Sinc(bDibD() As Byte, ByVal dstWidth&, ByVal dstHeight&, bDibS() As Byte, ByVal srcWidth&, ByVal srcHeight&)
    Dim x&, y&, X1&, Y1&, M&, N&, kX#, kY#, fX#, fY#
    Dim iR#, iG#, iB#, r1#, r2#
    kX = (srcWidth - 1) / (dstWidth - 1)
    kY = (srcHeight - 1) / (dstHeight - 1)
    For y = dstHeight - 1 To 0 Step -1
        fY = y * kY
        Y1 = Int(fY)
        fY = fY - Y1
        For x = 0 To dstWidth - 1
            fX = x * kX
            X1 = Int(fX)
            fX = fX - X1
            X1 = X1 * 3
            iR = 0: iG = 0: iB = 0
            ' Uses various kernel size
            For M = -kernel_size + 1 To kernel_size
                r1 = Windowed_Sinc(M - fY)
                For N = -kernel_size + 1 To kernel_size
                    r2 = Windowed_Sinc(fX - N)
                    iB = iB + bDibS(X1 + N * 3, Y1 + M) * r1 * r2
                    iG = iG + bDibS(X1 + N * 3 + 1, Y1 + M) * r1 * r2
                    iR = iR + bDibS(X1 + N * 3 + 2, Y1 + M) * r1 * r2
                Next
            Next
            If iB < 0 Then iB = 0
            If iG < 0 Then iG = 0
            If iR < 0 Then iR = 0
            If iB > 255 Then iB = 255
            If iG > 255 Then iG = 255
            If iR > 255 Then iR = 255
            bDibD(x * 3, y) = iB
            bDibD(x * 3 + 1, y) = iG
            bDibD(x * 3 + 2, y) = iR
        Next
        If frmResize.ShowProgress(dstHeight - y, dstHeight) = True Then Exit Sub
    Next
End Sub

' Normalization is useful especially for blurry interpolations
' like B-spline, Bell, Gaussian and some BC-spline with non-zero B parameter,
' e.g., filters with Y < 1 for X = 0 in Spatial Domain.
Sub Normalize(bDib() As Byte, ByVal nWidth&, ByVal nHeight&)
    Dim x&, y&, V As Byte, K&, H&(255)
    Dim Min As Byte, Max As Byte
    ' calculate histogram of interpolated image
    For y = nHeight - 1 To 0 Step -1
        For x = 0 To (nWidth - 1) * 3
            V = bDib(x, y)
            H(V) = H(V) + 1
        Next
        If frmResize.ShowProgress(nHeight - y, nHeight) = True Then Exit Sub
    Next
    ' Value 0.001 controls bounds of the histogram (tolerance).
    ' If it is zero then we have strict normalization
    ' but you often can't see any difference between
    ' original and normalized image.
    K = 0.001 * nWidth * nHeight
    Min = 96 ' the highest minimal value
    Max = 159 ' the lowest maximal value
    ' Get Min and Max values by given tolerance K
    For x = 0 To Min - 1
        If H(x) > K Then Min = x: Exit For
    Next
    For x = 255 To Max + 1 Step -1
        If H(x) > K Then Max = x: Exit For
    Next
    If Min = 0 And Max = 255 Then
        MsgBox "The image doesn't need normalization. [Aborted]", vbInformation
        Exit Sub
    End If
    ' Calculate lookup table with normalized values to accelerate processing
    For x = 0 To 255
        K = (x - Min) / (Max - Min) * 255
        If K < 0 Then K = 0
        If K > 255 Then K = 255
        H(x) = K
    Next
    ' Transform output image by lookup table
    For y = nHeight - 1 To 0 Step -1
        For x = 0 To (nWidth - 1) * 3
            bDib(x, y) = H(bDib(x, y))
        Next
        If frmResize.ShowProgress(nHeight - y, nHeight) = True Then Exit Sub
    Next
End Sub

Sub DoResample(ByVal nType As eTypes, hBitmapDst As Picture, hBitmapSrc As Picture)
    Dim bDibD() As Byte, bDibS() As Byte, bDibT() As Byte
    Dim tSAD As SAFEARRAY2D, tSAS As SAFEARRAY2D
    Dim tBMD As BITMAP, tBMS As BITMAP
    GetObjectAPI hBitmapDst, Len(tBMD), tBMD
    With tSAD ' Array header structure
        .cbElements = 1
        .cDims = 2
        .Bounds(0).cElements = tBMD.bmHeight
        .Bounds(1).cElements = tBMD.bmWidthBytes ' (Width*3 aligned to 4)
        .pvData = tBMD.bmBits ' Pointer to array (bitmap)
    End With
    ' Associate header with array (no need of copying large blocks, direct access)
    CopyMemory ByVal VarPtrArray(bDibD), VarPtr(tSAD), 4
    GetObjectAPI hBitmapSrc, Len(tBMS), tBMS
    With tSAS
        .cbElements = 1
        .cDims = 2
        .Bounds(0).cElements = tBMS.bmHeight
        .Bounds(1).cElements = tBMS.bmWidthBytes
        .pvData = tBMS.bmBits
    End With
    CopyMemory ByVal VarPtrArray(bDibS), VarPtr(tSAS), 4
    Select Case nType
        Case NearestNeighbor
            Resample_NearestNeighbor bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibS, tBMS.bmWidth, tBMS.bmHeight
        Case Bilinear
            ' Extend image by given size
            ReDim bDibT(-3 * (kernel_size - 1) To (tBMS.bmWidth - 1 + kernel_size) * 3 + 2, -(kernel_size - 1) To tBMS.bmHeight - 1 + kernel_size)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2 ' Copy source to extended with edge filling
            Resample_Bilinear bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
        Case BicubicCardinal
            ' Now we need 1 more row of points at top and left and 2 rows at right and bottom otherwise program will crash
            ReDim bDibT(-3 To tBMS.bmWidth * 3 + 5, -1 To tBMS.bmHeight + 1)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2
            Resample_BicubicCardinal bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
        Case BicubicBSpline
            ReDim bDibT(-3 To tBMS.bmWidth * 3 + 5, -1 To tBMS.bmHeight + 1)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2
            Resample_BicubicBSpline bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
        Case BicubicBCSpline
            ReDim bDibT(-3 To tBMS.bmWidth * 3 + 5, -1 To tBMS.bmHeight + 1)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2
            Resample_BicubicBCSpline bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
        Case Bell
            ReDim bDibT(-3 To tBMS.bmWidth * 3 + 5, -1 To tBMS.bmHeight + 1)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2
            Resample_Bell bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
        Case Gaussian
            ReDim bDibT(-3 * (kernel_size - 1) To (tBMS.bmWidth - 1 + kernel_size) * 3 + 2, -(kernel_size - 1) To tBMS.bmHeight - 1 + kernel_size)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2
            Resample_Gaussian bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
        Case WindowedSinc
            ReDim bDibT(-3 * (kernel_size - 1) To (tBMS.bmWidth - 1 + kernel_size) * 3 + 2, -(kernel_size - 1) To tBMS.bmHeight - 1 + kernel_size)
            CopyImage24 bDibS, bDibT, (tBMS.bmWidth - 1) * 3 + 2
            Resample_Sinc bDibD, tBMD.bmWidth, tBMD.bmHeight, bDibT, tBMS.bmWidth, tBMS.bmHeight
    End Select
    CopyMemory ByVal VarPtrArray(bDibD), 0&, 4 ' Important under WinNT platform
    CopyMemory ByVal VarPtrArray(bDibS), 0&, 4 ' Free arrays
End Sub


